"""Layer for small backward compatibility wrappers for DOLFINx"""

import dolfinx


def get_cmap(mesh: dolfinx.mesh.Mesh) -> dolfinx.fem.CoordinateElement:
    """Get the basix Cmap for the mesh."""
    if callable(mesh.geometry.cmap):
        return mesh.geometry.cmap()
    else:
        return mesh.geometry.cmap



from mpi4py import MPI
import dolfinx
import numpy as np
import numpy.typing as npt
from scipy.optimize import minimize
from scifem import closest_point_projection
from scifem.geometry import project_onto_simplex, _closest_point_projection
import ufl
import basix.ufl
import pytest


def scipy_project_point_to_element(
    mesh: dolfinx.mesh.Mesh,
    cells: npt.NDArray[np.int64],
    target_points: npt.NDArray[np.float64 | np.float32],
    method=None,
    tol: float | None = None,
):
    """
    Projects a 3D point onto a cell in a potentially higher order mesh.

    Args:
        mesh: {py:class}`dolfinx.mesh.Mesh`, the mesh containing the cell.
        cells: {py:class}`numpy.ndarray`, the indices of the cells to project onto.
        target_point: (3,) numpy array, the 3D point to project.
        method: str, the optimization method to use.
        tol: float, the tolerance for the optimizer.

    Returns:
        dict: A dictionary containing the reference coordinates, closest 3D point, and distance.
    """

    # Extract scalar element of mesh
    element = mesh.ufl_domain().ufl_coordinate_element().sub_elements[0]

    # Get the coordinates of the nodes for the specified cell
    node_coords = mesh.geometry.x[mesh.geometry.dofmap[cells]][:, :, : mesh.geometry.dim]
    target_points = target_points.reshape(-1, 3)
    # cmap = mesh.geometry.cmap

    # Constraints and Bounds
    cell_type = mesh.topology.cell_type
    if (
        cell_type == dolfinx.mesh.CellType.triangle
        or cell_type == dolfinx.mesh.CellType.tetrahedron
    ):
        method = method or "SLSQP"
        constraint = {"type": "ineq", "fun": lambda x: 1.0 - np.sum(x)}
        d = 1.0
    else:
        method = method or "L-BFGS-B"
        constraint = {}
        d = 0.0
    bounds = [(0.0, 1.0) for _ in range(mesh.topology.dim)]

    # Set initial guess and tolerance for solver

    initial_guess = np.full(mesh.topology.dim, 1 / (mesh.topology.dim + d), dtype=np.float64)
    tol = np.sqrt(np.finfo(mesh.geometry.x.dtype).eps) if tol is None else tol
    closest_points = np.zeros((target_points.shape[0], 3), dtype=mesh.geometry.x.dtype)
    ref_points = np.zeros((target_points.shape[0], mesh.topology.dim), dtype=mesh.geometry.x.dtype)
    for i, (coord, target_point) in enumerate(zip(node_coords, target_points)):
        coord = coord.reshape(-1, mesh.geometry.dim)

        def S(x_ref):
            N_vals = element.tabulate(0, x_ref.reshape(1, mesh.topology.dim))[0, 0, :]
            return np.dot(N_vals, coord)

        def dSdx_ref(x_ref):
            """Evaluate jacobian (tangent vectors) at the given reference coordinates."""
            dN = element.tabulate(1, x_ref.reshape(1, mesh.topology.dim))[
                1 : mesh.topology.dim + 1, 0, :
            ]
            return np.dot(dN, coord)

        def objective(x_ref):
            surface_point = S(x_ref)
            diff = surface_point - target_point
            return 0.5 * np.dot(diff, diff)

        def objective_grad(x_ref):
            diff = S(x_ref) - target_point
            tangents = dSdx_ref(x_ref)
            return np.dot(tangents, diff)

        res = minimize(
            objective,
            initial_guess,
            method=method,
            jac=objective_grad,
            bounds=bounds,
            constraints=constraint,
            tol=tol,
            options={"disp": False, "ftol": tol, "maxiter": 250},
        )
        closest_points[i] = S(res.x)
        ref_points[i] = res.x
        assert res.success, f"Optimization failed for {cells[i]} and {target_point=}: {res.message}"
    return closest_points, ref_points


@pytest.mark.parametrize("num_threads", [1, 2, 4])
@pytest.mark.parametrize("order", [1, 2])
def test_2D_manifold(order, num_threads):
    comm = MPI.COMM_SELF

    # Curved quadratic triangle in 3D (6 nodes)
    curved_nodes = np.array(
        [
            [0.0, 0.0, 0.0],  # Node 0: Vertex
            [1.0, 0.0, 0.0],  # Node 1: Vertex
            [0.0, 1.0, 0.0],  # Node 2: Vertex
            [0.6, 0.6, 0.0],  # Node 4: Edge 1-2
            [0.1, 0.5, 0.2],  # Node 5: Edge 2-0 (curved upward in Z)
            [0.5, 0.1, 0.2],  # Node 3: Edge 0-1 (curved upward in Z)
        ]
    )
    cells = np.array([[0, 1, 2, 3, 4, 5]], dtype=np.int64)  # Single curved triangle element
    c_el = ufl.Mesh(basix.ufl.element("Lagrange", "triangle", order, shape=(3,)))
    if order == 1:
        curved_nodes = curved_nodes[:3]  # Use only vertices for linear case
        cells = cells[:, :3]
    mesh = dolfinx.mesh.create_mesh(comm, cells=cells, x=curved_nodes, e=c_el)

    tol_x = 5e-6
    tol_dist = 1e-7
    theta = np.linspace(0, 4 * np.pi, 3_016)
    rand = np.random.RandomState(42)
    R = rand.rand(len(theta))
    z = rand.rand(len(theta)) * 0.5  # Add some random z variation
    points = np.vstack([R * np.cos(theta), R * np.sin(theta), z]).T

    cells = np.zeros(points.shape[0], dtype=np.int32)

    closest_point, closest_ref = closest_point_projection(
        mesh,
        cells,
        points,
        tol_x=tol_x,
        tol_dist=tol_dist,
        tol_grad=1e-16,
        max_iter=500,
        max_ls_iter=50,
        num_threads=num_threads,
    )

    (result_scipy, _ref_scipy) = scipy_project_point_to_element(mesh, cells, points, tol=tol_x**2)

    result, ref_coords = _closest_point_projection(
        mesh, cells, points, tol_x=tol_x, tol_dist=tol_dist
    )

    # Check that python and C++ implementations give the same result
    np.testing.assert_allclose(result, closest_point, atol=tol_dist)
    np.testing.assert_allclose(ref_coords, closest_ref, atol=10 * tol_x)

    # Check that scipy and our implementation give similar distances,
    # allowing for some tolerance due to different optimization methods
    diff_scifem = closest_point - points
    dist_scifem = 0.5 * np.sum(diff_scifem**2, axis=1)
    diff_scipy = result_scipy - points
    dist_scipy = 0.5 * np.sum(diff_scipy**2, axis=1)
    np.testing.assert_allclose(dist_scifem, dist_scipy, atol=50 * tol_dist, rtol=1e-6)

    for coord in closest_ref:
        # Check that we are within the bounds of the simplex
        ref_proj = project_onto_simplex(coord)
        np.testing.assert_allclose(ref_proj, coord)


@pytest.mark.parametrize("num_threads", [1, 2, 4])
@pytest.mark.parametrize("order", [1, 2])
def test_3D_curved_cell(order, num_threads):
    comm = MPI.COMM_SELF

    curved_nodes_tet = np.array(
        [
            [0.0, 0.0, 0.0],  # 0: Vertex
            [1.0, 0.0, 0.0],  # 1: Vertex
            [0.0, 1.0, 0.0],  # 2: Vertex
            [0.0, 0.0, 1.0],  # 3: Vertex
            [0.0, 0.5, 0.5],  # 4: Edge 2-3
            [0.5, 0.0, 0.5],  # 5: Edge 1-3
            [0.5, 0.5, 0.0],  # 6: Edge 1-2
            [0.0, 0.0, 0.5],  # 9: Edge 0-3
            [0.0, 0.5, 0.2],  # 8: Edge 0-2 (Curved)
            [0.5, 0.0, 0.2],  # 7: Edge 0-1 (Curved)
        ],
        dtype=np.float64,
    )
    cells_tet = np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]], dtype=np.int64)

    domain_tet = ufl.Mesh(basix.ufl.element("Lagrange", "tetrahedron", order, shape=(3,)))
    if order == 1:
        curved_nodes_tet = curved_nodes_tet[:4]  # Use only vertices for linear case
        cells_tet = cells_tet[:, :4]
    mesh = dolfinx.mesh.create_mesh(comm, cells=cells_tet, x=curved_nodes_tet, e=domain_tet)

    rand = np.random.RandomState(32)
    M = 5_001
    points = rand.rand(M, 3) - 0.5 * rand.rand(M, 3)
    tol_x = 1e-7
    tol_dist = 1e-7

    cells = np.zeros(points.shape[0], dtype=np.int32)
    closest_point, closest_ref = closest_point_projection(
        mesh,
        cells,
        points,
        tol_x=tol_x,
        tol_dist=tol_dist,
        tol_grad=1e-16,
        max_iter=500,
        max_ls_iter=50,
        num_threads=num_threads,
    )
    (result_scipy, _ref_scipy) = scipy_project_point_to_element(mesh, cells, points, tol=tol_x**2)
    result, ref_coords = _closest_point_projection(
        mesh,
        cells,
        points,
        tol_x=tol_x,
        tol_dist=tol_dist,
        tol_grad=1e-16,
    )

    # Check that python and C++ implementations give the same result
    np.testing.assert_allclose(result, closest_point, atol=tol_dist)
    np.testing.assert_allclose(ref_coords, closest_ref, atol=10 * tol_x)

    # Check that scipy and our implementation give similar distances,
    # allowing for some tolerance due to different optimization methods
    diff_scifem = closest_point - points
    dist_scifem = 0.5 * np.sum(diff_scifem**2, axis=1)
    diff_scipy = result_scipy - points
    dist_scipy = 0.5 * np.sum(diff_scipy**2, axis=1)
    np.testing.assert_allclose(dist_scifem, dist_scipy, atol=50 * tol_dist, rtol=1e-6)

    for coord in closest_ref:
        # Check that we are within the bounds of the simplex
        ref_proj = project_onto_simplex(coord)
        np.testing.assert_allclose(ref_proj, coord)



# Create a point source for Poisson problem
# Author: Jørgen S. Dokken
# SPDX-License-Identifier: MIT

from __future__ import annotations
from typing import TYPE_CHECKING

from packaging.version import Version
from mpi4py import MPI

if TYPE_CHECKING:
    from petsc4py import PETSc

try:
    from petsc4py import PETSc

    _HAS_PETSC = True
except ImportError:
    _HAS_PETSC = False

import dolfinx
import numpy as np
import numpy.typing as npt
import ufl

from .compat import get_cmap
from .utils import unroll_dofmap

__all__ = ["PointSource"]


class PointSource:
    """Class for defining a point source in a given function space."""

    def __init__(
        self,
        V: dolfinx.fem.FunctionSpace,
        points: npt.NDArray[np.float32] | npt.NDArray[np.float64],
        magnitude: np.floating | np.complexfloating = dolfinx.default_scalar_type(1),
        tol: np.floating | None = None,
    ) -> None:
        """Initialize a point source.

        Args:
            V: The function space the point source is defined in.
            points: The points where the point source is located.
                Input shape: ``(num_points, 3)``
            magnitude: The magnitudes of the point sources.
            tol: Tolerance for point location. If `None` 100 times machine epsilon is used.

        Note:
            Points should only be defined on one process. If they are sent in
            from multiple processes, multiple point sources will be created.

        Note:
            If the point source is outside the mesh, a ``ValueError`` will be raised.
        """
        self._function_space = V
        if V.dofmap.bs > 1 and dolfinx.__version__ == "0.8.0":
            raise NotImplementedError(
                "Block function spaces are not supported in dolfinx 0.8.0. Please upgrade dolfinx"
            )
        if points.ndim == 1:
            points = points.reshape(1, -3)
        elif points.ndim == 2 and points.shape[1] != 3:
            raise ValueError(f"Points should have shape (num_points, 3), got {points.shape}")

        self._input_points = points
        self._magnitude = magnitude
        # Initialize empty arrays
        self._points = np.empty((0, 3), dtype=points.dtype)
        self._cells = np.empty(0, dtype=np.int32)
        num_dofs = self._function_space.dofmap.dof_layout.num_dofs * self._function_space.dofmap.bs
        self._basis_values = np.empty(
            (0, num_dofs), dtype=self._function_space.mesh.geometry.x.dtype
        )
        self._tol = (
            float(1e2 * np.finfo(self._input_points.dtype).eps) if tol is None else float(tol)
        )
        self.recompute_sources()
        self.compute_cell_contributions()

    def recompute_sources(self, tol: float | None = None):
        """Recompute the what cells the point sources collide with.

        This function should be called if the mesh geometry has been modified.

        Args:
            tol: Tolerance for point location. If `None`, the tolerance from initialization is used.
        """

        # Determine what process owns a point and what cells it lies within
        mesh = self._function_space.mesh
        tol = self._tol if tol is None else tol
        if dolfinx.__version__ == "0.8.0":
            src_ranks, _, self._points, self._cells = (
                dolfinx.cpp.geometry.determine_point_ownership(
                    mesh._cpp_object, self._input_points, tol
                )
            )
            self._points = np.array(self._points).reshape(-1, 3)
        elif Version(dolfinx.__version__) >= Version("0.9.0.0"):
            try:
                collision_data = dolfinx.cpp.geometry.determine_point_ownership(
                    mesh._cpp_object, self._input_points, tol
                )
            except TypeError:
                collision_data = dolfinx.geometry.determine_point_ownership(
                    mesh, self._input_points, padding=tol
                )

            self._points = collision_data.dest_points
            self._cells = collision_data.dest_cells
            src_ranks = collision_data.src_owner
        else:
            raise NotImplementedError(f"Unsupported version of dolfinx: {dolfinx.__version__}")
        if -1 in src_ranks:
            raise ValueError("Point source is outside the mesh.")

    def compute_cell_contributions(self):
        """Compute the basis function values at the point sources."""
        mesh = self._function_space.mesh
        # Pull owning points back to reference cell
        mesh_nodes = mesh.geometry.x
        cmap = get_cmap(mesh)

        ref_x = np.zeros((len(self._cells), mesh.topology.dim), dtype=mesh.geometry.x.dtype)
        for i, (point, cell) in enumerate(zip(self._points, self._cells)):
            geom_dofs = mesh.geometry.dofmap[cell]
            ref_x[i] = cmap.pull_back(point.reshape(-1, 3), mesh_nodes[geom_dofs])

        # Create expression evaluating a trial function (i.e. just the basis function)
        u = ufl.TestFunction(self._function_space)
        bs = self._function_space.dofmap.bs
        num_dofs = self._function_space.dofmap.dof_layout.num_dofs * bs
        if len(self._cells) > 0:
            # NOTE: Expression lives on only this communicator rank
            # Expression is evaluated for every point in every cell, which means that we
            # need to discard values that are not on the diagonal.
            expr = dolfinx.fem.Expression(u, ref_x, comm=MPI.COMM_SELF)
            all_values = expr.eval(mesh, self._cells)
            # Diagonalize values (num_cells, num_points, num_dofs, bs) -> (num_cells, num_dofs)
            if Version(dolfinx.__version__) <= Version("0.9.0"):
                basis_values = np.empty((len(self._cells), num_dofs), dtype=all_values.dtype)
                for i in range(len(self._cells)):
                    basis_values[i] = sum(
                        [
                            all_values[i, i * num_dofs * bs : (i + 1) * num_dofs * bs][
                                j * num_dofs : (j + 1) * num_dofs
                            ]
                            for j in range(bs)
                        ]
                    )
            else:
                basis_values = np.zeros(
                    (len(self._cells), num_dofs), dtype=dolfinx.default_scalar_type
                )
                if bs == 1:
                    # Values have shape (num_cells, num_points, num_dofs)
                    for i in range(len(self._cells)):
                        basis_values[i] = all_values[i, i, :]
                else:
                    # Values have shape (num_cells, num_points, bs, num_dofs*bs)
                    for i in range(len(self._cells)):
                        basis_values[i] = sum(all_values[i, i, j, :] for j in range(bs))
        else:
            basis_values = np.zeros((0, num_dofs), dtype=dolfinx.default_scalar_type)
        self._basis_values = basis_values

    def apply_to_vector(
        self, b: dolfinx.fem.Function | dolfinx.la.Vector | "PETSc.Vec", recompute: bool = False
    ):
        """Apply the point sources to a vector.

        Args:
            b: The vector to apply the point sources to.
            recompute: If the point sources should be recomputed before applying.
                Recomputation should only be done if the mesh geometry has been modified.

        Note:
            The user is responsible for forward scattering of the vector after
            applying the point sources.

        Note:
            If a PETSc vector is passed in, one has to call ``b.assemble()`` prior to solving the
            linear system (post scattering).
        """
        if recompute:
            self.recompute_sources()
            self.compute_cell_contributions()

        # Apply the point sources to the vector
        _dofs = self._function_space.dofmap.list[self._cells]
        unrolled_dofs = unroll_dofmap(_dofs, self._function_space.dofmap.bs)
        for dofs, values in zip(unrolled_dofs, self._basis_values, strict=True):
            if isinstance(b, dolfinx.fem.Function):
                b.x.array[dofs] += values * self._magnitude
            elif isinstance(b, dolfinx.la.Vector):
                b.array[dofs] += values * self._magnitude
            elif _HAS_PETSC and isinstance(b, PETSc.Vec):
                b.setValuesLocal(
                    dofs,
                    values * self._magnitude,
                    addv=PETSc.InsertMode.ADD_VALUES,
                )
            else:
                raise TypeError(f"Unsupported vector type {type(b)}.")



import pytest
import builtins
import basix.ufl
from petsc4py import PETSc
from mpi4py import MPI
import numpy as np
import dolfinx
from scifem import evaluate_function, compute_extrema
import ufl


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
def test_evaluate_scalar_function_2D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = 2.0
    nx = ny = 10

    mesh = dolfinx.mesh.create_rectangle(
        comm=comm,
        points=[np.array([0.0, 0.0]), np.array([Lx, Ly])],
        n=[nx, ny],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: x[0] + 2 * x[1]
    u.interpolate(f)

    points = np.array([[0.0, 0.0], [0.2, 0.2], [0.5, 0.5], [0.7, 0.2]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values[:, 0], exact)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
def test_evaluate_vector_function_2D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = 2.0
    nx = ny = 10

    mesh = dolfinx.mesh.create_rectangle(
        comm=comm,
        points=[np.array([0.0, 0.0]), np.array([Lx, Ly])],
        n=[nx, ny],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1, (2,)))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: (x[0], 2 * x[1])
    u.interpolate(f)

    points = np.array([[0.0, 0.0], [0.2, 0.2], [0.5, 0.5], [0.7, 0.2]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values, exact)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
def test_evaluate_scalar_function_3D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = Lz = 2.0
    nx = ny = nz = 10

    mesh = dolfinx.mesh.create_box(
        comm=comm,
        points=[np.array([0.0, 0.0, 0.0]), np.array([Lx, Ly, Lz])],
        n=[nx, ny, nz],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: x[0] + 2 * x[1]
    u.interpolate(f)

    points = np.array([[0.0, 0.0, 0.0], [0.2, 0.2, 0.3], [0.5, 0.5, 0.2], [0.7, 0.2, 0.5]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values[:, 0], exact)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
def test_evaluate_vector_function_3D(cell_type):
    comm = MPI.COMM_WORLD
    Lx = Ly = Lz = 2.0
    nx = ny = nz = 10

    mesh = dolfinx.mesh.create_box(
        comm=comm,
        points=[np.array([0.0, 0.0, 0.0]), np.array([Lx, Ly, Lz])],
        n=[nx, ny, nz],
        cell_type=cell_type,
    )

    V = dolfinx.fem.functionspace(mesh, ("P", 1, (3,)))
    u = dolfinx.fem.Function(V, dtype=PETSc.ScalarType)
    f = lambda x: (x[0], 2 * x[1], -x[2])
    u.interpolate(f)

    points = np.array([[0.0, 0.0, 0.0], [0.2, 0.2, 0.3], [0.5, 0.5, 0.2], [0.7, 0.2, 0.5]])
    u_values = evaluate_function(u, points)
    exact = np.array(f(points.T)).T

    assert np.allclose(u_values, exact)


@pytest.mark.parametrize("degree", [5])
@pytest.mark.parametrize("extrema", [min, max])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
@pytest.mark.parametrize("dtype", [np.float64, np.float32])
def test_extrema_func(cell_type, extrema, dtype, degree: int):
    tol = 100 * np.finfo(dtype).eps

    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 6, 6, 6, cell_type=cell_type, dtype=dtype)
    el = basix.ufl.element("Lagrange", mesh.basix_cell(), degree, dtype=dtype)
    V = dolfinx.fem.functionspace(mesh, el)

    x_p = np.array([0.318, 0.852, 0.53], dtype=dtype)

    sign = 1 if extrema is builtins.min else -1

    def f(x):
        return (
            -0.8
            * sign
            * np.exp(-sum([(x[i] - x_p[i]) ** 2 for i in range(mesh.geometry.dim)]) / 0.1)
        )

    u = dolfinx.fem.Function(V, dtype=dtype)
    u.interpolate(f)

    u_ex, _X_ex = compute_extrema(u, extrema, tol=tol)

    assert np.isclose(u_ex, -sign * 0.8, atol=tol)

    x_ufl = ufl.SpatialCoordinate(mesh)
    x_p_ufl = ufl.as_vector(x_p)
    f_ufl = -0.8 * sign * ufl.exp(-ufl.dot(x_ufl - x_p_ufl, x_ufl - x_p_ufl) / 0.1)

    u_ufl_ex, _X_ufl_ex = compute_extrema(f_ufl, extrema, tol=tol)

    assert np.isclose(u_ufl_ex, -sign * 0.8, atol=tol)
    np.testing.assert_allclose(_X_ufl_ex, x_p, atol=tol)



import dolfinx
import numpy as np
import numpy.typing as npt
import warnings
from scifem._scifem import closest_point_projection_float64, closest_point_projection_float32


def project_onto_simplex(
    v: npt.NDArray[np.float64 | np.float32],
) -> npt.NDArray[np.float64 | np.float32]:
    """
    Exact projection of vector v onto the simplex {x >= 0, sum(x) <= 1}.

    See Algorithm 1: Laurent Condat. Fast Projection onto the Simplex and the l1 Ball.
    Mathematical Programming, Series A, 2016, 158 (1), pp.575-585.
    ⟨DOI: 10.1007/s10107-015-0946-6⟩.


    Args:
        v: The vector to project onto simplex

    Return:
        The projection of v onto the simplex.

    """
    # 1. First try the unconstrained positive quadrant projection
    v = v.reshape(-1)
    u = np.maximum(v, 0.0)
    if np.sum(u) <= 1.0:
        return u

    # 2. Otherwise, project exactly onto the slanted face sum(x) = 1
    tdim = len(v)
    sort_v = np.sort(v)[::-1]
    cssv = np.cumsum(sort_v)

    # Find the primal-dual root
    # sum x_i = a
    # Find K: = max (k in [1,.. N] such that sum_{r=1}^k sort_v[r] - a)/k < u_k
    # Multiply by k and take the last true entry to get the index rho
    K = np.nonzero(sort_v * np.arange(1, tdim + 1) > (cssv - 1.0))[0][-1]
    tau = (cssv[K] - 1.0) / (K + 1.0)

    return np.maximum(v - tau, 0.0)


def closest_point_projection(
    mesh: dolfinx.mesh.Mesh,
    cells: npt.NDArray[np.int32],
    target_points: npt.NDArray[np.float64 | np.float32],
    tol_x: float | None = None,
    tol_dist: float = 1e-10,
    tol_grad: float = 1e-10,
    max_iter: int = 2000,
    max_ls_iter: int = 250,
    num_threads: int = 1,
):
    """Projects a 3D point onto a cell in a potentially higher order mesh.

    Uses the Goldstein-Levitin-Polyak Gradient projection method, where
    potential simplex constraints are handled by an exact projection using a
    primal-dual root finding method. See:
    - Held, M., Wolfe, P., Crowder, H.: Validation of subgradient optimization (1974)
    - Laurent Condat. Fast Projection onto the Simplex and the l1 Ball. (2016)
    - Dimitri P. Bertsekas, "On the Goldstein-Levitin-Polyak gradient projection method," (1976)

    Args:
        mesh: The mesh containing the cells.
        cells: The local indices of the cells to project onto.
        target_points: (n, 3) numpy array, the 3D points to project.
        tol_x: Tolerance for changes between iterates in the reference coordinates.
            If None, uses the square root of machine precision.
        tol_grad: Tolerance for determination based on the gradient norm.
            The gradient is scaled by the Jacobian to account for stretching.
        tol_dist: Tolerance used to determine if the projected point is close enough to
            the target point to stop optimization.
        max_iter: int, the maximum number of iterations for the projected gradient method.
        max_ls_iter: int, the maximum number of line search iterations.
        num_threads: int, the number of threads to use for parallel projection.

    Returns:
        A tuple of arrays containing the closest points (in physical space)
        and reference coordinates for each cell to each target point.
    """
    xdtype = mesh.geometry.x.dtype

    if xdtype == np.float64:
        return closest_point_projection_float64(
            mesh._cpp_object,
            cells,
            target_points,
            tol_x=tol_x,
            tol_dist=tol_dist,
            tol_grad=tol_grad,
            max_iter=max_iter,
            max_ls_iter=max_ls_iter,
            num_threads=num_threads,
        )
    elif xdtype == np.float32:
        return closest_point_projection_float32(
            mesh._cpp_object,
            cells,
            target_points,
            tol_x=tol_x,
            tol_dist=tol_dist,
            tol_grad=tol_grad,
            max_iter=max_iter,
            max_ls_iter=max_ls_iter,
            num_threads=num_threads,
        )
    else:
        raise TypeError(f"Unsupported mesh coordinate dtype {xdtype}.")


def _closest_point_projection(
    mesh: dolfinx.mesh.Mesh,
    cells: npt.NDArray[np.int32],
    target_points: npt.NDArray[np.float64 | np.float32],
    tol_x: float | None = None,
    tol_dist: float = 1e-10,
    tol_grad: float = 1e-10,
    max_iter: int = 2000,
    max_ls_iter: int = 250,
) -> tuple[npt.NDArray[np.float64 | np.float32], npt.NDArray[np.float64 | np.float32]]:
    """Projects a 3D point onto a cell in a potentially higher order mesh.

    NOTE:
        Reference implementation in Python. This should only be used for testing and debugging,
        and is not optimized for performance. The performant implementation can be found in
        {py:func}`closest_point_projection`.

    Uses the Goldstein-Levitin-Polyak Gradient projection method, where
    potential simplex constraints are handled by an exact projection using a
    primal-dual root finding method. See:
    - Held, M., Wolfe, P., Crowder, H.: Validation of subgradient optimization (1974)
    - Laurent Condat. Fast Projection onto the Simplex and the l1 Ball. (2016)
    - Dimitri P. Bertsekas, "On the Goldstein-Levitin-Polyak gradient projection method," (1976)

    Args:
        mesh: The mesh containing the cell.
        cells: The local indices of the cells to project onto.
        target_points: (n, 3) numpy array, the 3D points to project.
        tol_x: Tolerance for changes between iterates in the reference coordinates.
            If None, uses the square root of machine precision.
        tol_grad: Tolerance for determination based on the gradient norm.
            The gradient is scaled by the Jacobian to account for stretching.
        tol_dist: Tolerance used to determine if the projected point is close enough to
            the target point to stop optimization.
        max_iter: int, the maximum number of iterations for the projected gradient method.
        max_ls_iter: int, the maximum number of line search iterations.

    Returns:
        A tuple of arrays containing the closest points (in physical space)
        and reference coordinates for each cell to each target point.
    """
    dtype = mesh.geometry.x.dtype
    eps = np.finfo(dtype).eps
    tol_x = np.sqrt(eps) if tol_x is None else tol_x
    roundoff_tol = 100 * eps

    # Extract scalar element of mesh
    element = mesh.ufl_domain().ufl_coordinate_element().sub_elements[0]
    tdim = mesh.topology.dim
    # Get the coordinates of the nodes for the specified cell
    node_coords = mesh.geometry.x[mesh.geometry.dofmap[cells]][:, :, : mesh.geometry.dim]
    target_points = target_points.reshape(-1, 3)

    # Constraints and Bounds
    cell_type = mesh.topology.cell_type

    # Set initial guess and tolerance for solver
    if (
        cell_type == dolfinx.mesh.CellType.triangle
        or cell_type == dolfinx.mesh.CellType.tetrahedron
    ):
        plus_val = 1
    else:
        plus_val = 0
    initial_guess = np.full(mesh.topology.dim, 1 / (mesh.topology.dim + plus_val), dtype=dtype)
    closest_points = np.zeros((target_points.shape[0], 3), dtype=dtype)
    reference_points = np.zeros((target_points.shape[0], mesh.topology.dim), dtype=dtype)
    is_simplex = cell_type in [
        dolfinx.mesh.CellType.triangle,
        dolfinx.mesh.CellType.tetrahedron,
    ]

    if is_simplex:

        def project(x):
            return project_onto_simplex(x)
    else:

        def project(x):
            return np.clip(x, 0.0, 1.0)

    for i, (coord, target_point) in enumerate(zip(node_coords, target_points)):
        target_point = target_point[: mesh.geometry.dim]
        coord = coord.reshape(-1, mesh.geometry.dim)
        x_k = initial_guess.copy()

        for k in range(max_iter):
            x_old = x_k.copy()

            # Evaluate basis functions and first order derivatives at current reference point
            tab = element.tabulate(1, x_k.reshape(1, tdim))

            # Push forward to physical space
            surface_point = np.dot(tab[0, 0, :], coord)

            # Compute current objective function
            diff = surface_point - target_point
            current_dist_sq = 0.5 * np.dot(diff, diff)
            # Compute the gradient in reference coordinates using the Jacobian (tangent vectors)
            tangents = np.dot(tab[1 : tdim + 1, 0, :], coord)
            g = np.dot(tangents, diff)

            # Check for convergence in gradient norm, scaled by the Jacobian to account
            # for stretching of the reference space
            jac_norm = np.linalg.norm(tangents)

            scaled_tol_grad = tol_grad * max(jac_norm, 1.0)
            if np.linalg.norm(g) < scaled_tol_grad:
                break

            # 3. Goldstein-Polyak-Levitin Projected Line Search
            # Bertsekas (1976) Eq. (14) - Armijo Rule along the Projection Arc
            sigma = 0.1  # Sufficient decrease parameter (0 < sigma < 0.5)
            beta = 0.5  # Reduction factor (0 < beta < 1)
            alpha = 1.0  # Initial step size

            x_new_prev = np.full(tdim, -1, dtype=dtype)
            target_reached = False
            for li in range(max_ls_iter):
                # Apply the exact analytical simplex projection
                x_new = project(x_k - alpha * g)

                if np.linalg.norm(x_new - x_new_prev) < eps:
                    # The projection is pinned to a boundary.
                    # Changing alpha further will not change the physical point!
                    break
                x_new_prev = x_new.copy()

                # The actual physical step we took after hitting the geometric walls
                actual_step = x_new - x_k

                # Evaluate distance at the projected point
                tab_new = element.tabulate(0, x_new.reshape(1, tdim))
                S_new = np.dot(tab_new[0, 0, :], coord)
                new_dist_sq = 0.5 * np.dot(S_new - target_point, S_new - target_point)
                if new_dist_sq < 0.5 * tol_dist**2:
                    # We are close enough to the target point, no need for further line search
                    target_reached = True
                    break

                # Bertsekas Eq. (14) condition:
                # f(x_new) <= f(x_k) + sigma * grad_f(x_k)^T * (x_new - x_k)
                # Note: g is grad_f(x_k)
                if new_dist_sq <= current_dist_sq + sigma * np.dot(g, actual_step) + roundoff_tol:
                    # Condition satisfied
                    break

                # Reduction step (Backtracking)
                alpha *= beta

            if li == max_ls_iter - 1:
                warnings.warn(
                    f"Line search failed to converge after {max_ls_iter} iterations "
                    + f"for cell {cells[i]}  and {target_point=}."
                )
            x_k[:] = x_new

            if target_reached:
                break

            # 4. Check for convergence
            if np.linalg.norm(x_k - x_old) < tol_x:
                break

        assert np.allclose(project(x_k), x_k), "Projection failed to satisfy constraints"

        # Final coordinate extraction
        tab_final = element.tabulate(0, x_k.reshape(1, tdim))
        closest_points[i] = np.dot(tab_final[0, 0, :], coord)
        reference_points[i] = x_k

        if k == max_iter - 1:
            raise RuntimeError(
                f"Projected gradient method failed to converge after {max_iter} iterations ",
                f"for cell {cells[i]} and {target_point=} and final iterate {x_k=} ",
                f"and final point {closest_points[i]} with final distance ",
                f"{np.linalg.norm(closest_points[i] - target_point)}.",
            )
    return closest_points, reference_points



from mpi4py import MPI
import dolfinx
import numpy as np
import numpy.typing as npt
import builtins
import typing
import tempfile

import basix
import ufl
from scipy.optimize import minimize
from ufl.algorithms.signature import compute_expression_signature
from .compat import get_cmap


T = typing.TypeVar("T", int, float)
MinMaxFunc = typing.Callable[[typing.Sequence[T]], T]


__all__ = ["evaluate_function", "find_cell_extrema", "compute_extrema"]


def _compute_expression_signature(expr: ufl.core.expr.Expr) -> str:
    coeffs = ufl.algorithms.extract_coefficients(expr)
    consts = ufl.algorithms.analysis.extract_constants(expr)
    args = ufl.algorithms.analysis.extract_arguments(expr)
    assert len(args) == 0

    rn = dict()
    rn.update(dict((c, i) for i, c in enumerate(coeffs)))
    rn.update(dict((c, i) for i, c in enumerate(consts)))

    domains: list[ufl.AbstractDomain] = []
    for coeff in coeffs:
        domains.append(*ufl.domain.extract_domains(coeff))
    for gc in ufl.algorithms.analysis.extract_type(expr, ufl.classes.GeometricQuantity):
        domains.append(*ufl.domain.extract_domains(gc))
    for const in consts:
        domains.append(*ufl.domain.extract_domains(const))
    domains = ufl.algorithms.analysis.unique_tuple(domains)
    assert all([isinstance(domain, ufl.Mesh) for domain in domains])
    rn.update(dict((d, i) for i, d in enumerate(domains)))
    return compute_expression_signature(expr, rn)


def evaluate_function(
    u: dolfinx.fem.Function, points: npt.ArrayLike, broadcast=True
) -> npt.NDArray[np.float64]:
    """Evaluate a function at a set of points.

    Args:
        u: The function to evaluate.
        points: The points to evaluate the function at.
        broadcast: If True, the values will be broadcasted to all processes.

            Note:
                Uses a global MPI call to broadcast values, thus this has to
                be called on all active processes synchronously.

            Note:
                If the function is discontinuous, different processes may return
                different values for the same point.
                In this case, the value returned is the maximum value across all processes.

    Returns:
        The values of the function evaluated at the points.


    """
    mesh = u.function_space.mesh
    u.x.scatter_forward()
    comm = mesh.comm
    points = np.array(points, dtype=np.float64)
    assert len(points.shape) == 2, (
        f"Expected points to have shape (num_points, dim), got {points.shape}"
    )
    num_points = points.shape[0]
    extra_dim = 3 - mesh.geometry.dim

    # Append zeros to points if the mesh is not 3D
    if extra_dim > 0:
        points = np.hstack((points, np.zeros((points.shape[0], extra_dim))))

    bb_tree = dolfinx.geometry.bb_tree(mesh, mesh.topology.dim)
    # Find cells whose bounding-box collide with the the points
    potential_colliding_cells = dolfinx.geometry.compute_collisions_points(bb_tree, points)
    # Choose one of the cells that contains the point
    adj = dolfinx.geometry.compute_colliding_cells(mesh, potential_colliding_cells, points)
    indices = np.flatnonzero(adj.offsets[1:] - adj.offsets[:-1])
    cells = adj.array[adj.offsets[indices]]
    points_on_proc = points[indices]

    values = u.eval(points_on_proc, cells)
    if broadcast:
        bs = u.function_space.dofmap.index_map_bs
        # Create array to store values and fill with -inf
        # to ensure that all points are included in the allreduce
        # with op=MPI.MAX
        u_out = np.ones((num_points, bs), dtype=np.float64) * -np.inf
        # Fill in values for points on this process
        u_out[indices, :] = values
        # Now loop over all processes and find the maximum value
        for i in range(num_points):
            if bs > 1:
                # If block size is larger than 1, loop over blocks
                for j in range(bs):
                    u_out[i, j] = comm.allreduce(u_out[i, j], op=MPI.MAX)
            else:
                u_out[i] = comm.allreduce(u_out[i], op=MPI.MAX)

        return u_out
    else:
        return values


def find_cell_extrema(
    u: ufl.core.expr.Expr,
    cell: int,
    kind: MinMaxFunc,
    x0: npt.NDArray | None = None,
    jit_options: dict | None = None,
    method: str | None = None,
    options: dict | None = None,
    tol: float | None = None,
) -> tuple[npt.NDArray[np.floating], np.floating]:
    """
    Find the extrema of a {py:class}`ufl.core.expr.Expr` within a cell.

    Args:
        u: The expression to find the extrema of.
        cell: The local index of the cell to search in
        kind: If we search for minima or maxima
        x_0: The point to start the initial search at
        method: Optimization algorithm to use for local problem
        options: Options for optimization method
        tol: Tolerance for scipy minimize

    Returns:
        The point (in physical space) of the extrema and the value at the extrema.
    """
    if kind is builtins.min:
        sign = 1
    elif kind is builtins.max:
        sign = -1
    else:
        raise NotImplementedError(f"Unknown type of extrema ({kind}), expected {min} or {max}")

    ud = u.ufl_domain()
    assert ud is not None
    mesh = dolfinx.mesh.Mesh(ud.ufl_cargo(), ud)

    if x0 is None:
        x_ref = np.zeros(mesh.topology.dim, dtype=mesh.geometry.x.dtype)
    else:
        x_ref = x0

    _cell = np.array([cell], dtype=np.int32)
    mesh_nodes = mesh.geometry.x[mesh.geometry.dofmap[cell], : mesh.geometry.dim]
    _x_p = np.zeros(3)
    cmap = get_cmap(mesh)

    def eval_J(x_ref):
        # Evaluating basis functions through {py:func}`dolfinx.fem.Function.eval`
        # is faster than generating an expression for the same thing
        if isinstance(u, dolfinx.fem.Function):
            # This could in theory be made even faster by taking out some of the eval code
            # However, quite a lot of work needs to be reimplemented for minimal gain
            # to do so, so we rather push forward, then let eval pull back again.

            _x_p[: mesh.geometry.dim] = cmap.push_forward(
                x_ref.reshape(-1, mesh.topology.dim), mesh_nodes
            )[0]
            try:
                u_eval = u.eval(_x_p, _cell)[0]
            except RuntimeError:
                # Nonlinear pullback might fail on low precision
                u_expr = dolfinx.fem.Expression(
                    u,
                    x_ref,
                    comm=MPI.COMM_SELF,
                    jit_options=jit_options,
                    dtype=mesh.geometry.x.dtype,
                )
                u_eval = u_expr.eval(mesh, _cell)[0][0]
        else:
            # Expression has to be used for any UFL-expression that is not a Function
            u_expr = dolfinx.fem.Expression(
                u, x_ref, comm=MPI.COMM_SELF, jit_options=jit_options, dtype=mesh.geometry.x.dtype
            )
            u_eval = u_expr.eval(mesh, _cell)[0][0]
        return np.float64(sign * u_eval)  # SLSQP only supports float64

    def eval_dJ(x_ref):
        # Could be improved if we could take `ufl.classes.ReferenceGrad(u)`
        # https://github.com/FEniCS/ufl/issues/450
        u_grad_expr = dolfinx.fem.Expression(
            ufl.Jacobian(mesh).T * ufl.grad(u),
            x_ref,
            comm=MPI.COMM_SELF,
            jit_options=jit_options,
            dtype=mesh.geometry.x.dtype,
        )
        u_grad_eval = u_grad_expr.eval(mesh, _cell)[0][0]
        return (sign * u_grad_eval).astype(np.float64)  # SLSQP only supports float64

    # Bounds force x and y to be between 0 and 1
    bounds = [(0.0, 1.0) for _ in range(mesh.topology.dim)]

    # In SciPy, 'ineq' means the function must be >= 0.
    # So, x + y <= 1 becomes 1 - x - y >= 0.
    cell_type = mesh.topology.cell_type
    if cell_type == dolfinx.mesh.CellType.triangle:
        method = method or "SLSQP"
        constraint = {"type": "ineq", "fun": lambda x: 1.0 - x[0] - x[1]}
    elif (
        cell_type == dolfinx.mesh.CellType.quadrilateral
        or cell_type == dolfinx.mesh.CellType.hexahedron
    ):
        method = method or "L-BFGS-B"
        constraint = {}
    elif cell_type == dolfinx.mesh.CellType.tetrahedron:
        method = method or "SLSQP"
        constraint = {"type": "ineq", "fun": lambda x: 1.0 - x[0] - x[1] - x[2]}
    else:
        raise RuntimeError(f"Unsupported {cell_type=}")

    # --- 3. Run Optimization ---
    tol = tol if tol is not None else 100 * np.finfo(mesh.geometry.x.dtype).eps
    result = minimize(
        fun=eval_J,
        x0=x_ref.flatten(),
        method=method,
        jac=eval_dJ,
        bounds=bounds,
        constraints=constraint,
        options=options or {},
        tol=tol,
    )

    X_phys = cmap.push_forward(result.x.reshape(-1, mesh.topology.dim), mesh_nodes)[0]
    return X_phys, sign * result.fun


def compute_LP_average(
    u: ufl.core.expr.Expr, p: int, domain: dolfinx.mesh.Mesh
) -> npt.NDArray[np.float32 | np.float64 | np.complex128 | np.complex64]:
    r"""
    Compute the :math:`L^p(\Omega)`-average of a scalar-valued
    {py:class}`ufl-expression<ufl.core.expr.Expr>` over each cell in a
    {py:class}`domain<dolfinx.mesh.Mesh>`

    Args:
        u: The UFL-expression to evaluate
        p: The degree of the norm to use.
        domain: The integration domain

    Returns:
        An array of the local average L^P norm per local cell on the process,
        including ghosts.
    """
    if u.ufl_shape != ():
        raise ValueError(f"u must be scalar-valued, got {u.ufl_shape}.")
    V = dolfinx.fem.functionspace(domain, ("DG", 0))
    v = ufl.TestFunction(V)
    L_p = dolfinx.fem.assemble_vector(
        dolfinx.fem.form(u**p * ufl.conj(v) * ufl.dx, dtype=domain.geometry.x.dtype)
    )
    L_p.scatter_reverse(dolfinx.la.InsertMode.add)
    L_p.scatter_forward()
    cell_vol = dolfinx.fem.assemble_vector(
        dolfinx.fem.form(ufl.conj(v) * ufl.dx, dtype=domain.geometry.x.dtype)
    )
    cell_vol.scatter_reverse(dolfinx.la.InsertMode.add)
    cell_vol.scatter_forward()
    avg = L_p.array / cell_vol.array
    return avg


def compute_extrema(
    u: ufl.core.expr.Expr,
    kind=MinMaxFunc,
    p: int = 1,
    num_candidates: int = 3,
    x0: npt.NDArray[np.floating] | None = None,
    method: str | None = None,
    options: dict | None = None,
    tol: float | None = None,
    jit_options: dict[str, typing.Any] | None = None,
) -> tuple[np.floating, npt.NDArray[np.floating]]:
    """
    Find the extrema of an expression across its integration domain.

    Args:
        u: The expression
        p: Integer to compute the cell-wise L^p norm to speed up search.
        kind: `min` or `max`.
        num_candidates: The number of candidates with the largert average $L^p$
            norm over a cell.
        x0: Initial point in reference cell to start search at.
        method: Optimization algorithm to use for local problem
        options: Options for optimization method
        tol: Tolerance for scipy minimize

    Returns:
        The value at the extrema and the physical point
    """

    # Extract DOLFINx mesh
    ud = u.ufl_domain()
    assert ud is not None
    mesh = dolfinx.mesh.Mesh(ud.ufl_cargo(), ud)

    if kind is builtins.max:
        sign = -1
    elif kind is builtins.min:
        sign = 1
    else:
        raise ValueError("Unknown extrema {kind}, supports min and max only.")

    # Compute cell candidates
    local_avg = compute_LP_average(u, p, mesh)
    order = np.argsort(local_avg)
    if kind is builtins.max:
        candidate_cells = order[-num_candidates:]
    else:
        candidate_cells = order[:num_candidates]

    # Get reference point as midpoint
    if x0 is None:
        vertices = basix.geometry(mesh.basix_cell())
        x0 = np.average(vertices, axis=0)

    # Set max efficiency for temporary expression code.
    jit_options = (
        jit_options.copy()
        if jit_options is not None
        else {
            "cffi_extra_compile_args": ["-march=native", "-O3"],
            "cffi_libraries": ["m"],
        }
    )
    # Set temporary cache dir
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=False, delete=True) as cache_dir:
        jit_options["cache_dir"] = cache_dir
        # Find local extrema
        local_min = np.inf
        local_X = None
        for cell in candidate_cells:
            X_c, extrema = find_cell_extrema(
                u,
                cell,
                kind=kind,
                x0=x0,
                jit_options=jit_options,
                method=method,
                options=options,
                tol=tol,
            )
            if sign * extrema < local_min:
                local_min = sign * extrema
                local_X = X_c

    # Find global extrema
    val, rank = mesh.comm.allreduce((local_min, mesh.comm.rank), op=MPI.MINLOC)
    X = mesh.comm.bcast(local_X, root=rank)
    return sign * val, X
